"""Generate python files to access the stiebel eltron heat pumps."""

from pathlib import Path
import csv

from dataclasses import dataclass

from jinja2 import Environment, FileSystemLoader, Template, select_autoescape
from pystiebeleltron import RegisterType

ENERGY_DATA_BLOCK_NAME = "Energy Data"


@dataclass
class ModbusFile:
    """Register file class."""

    name: str
    path: Path
    register_type: RegisterType


def python_name(name: str, suffix: str = "") -> str:
    """Generate a valid python variable name."""
    if suffix == "":
        result = name.strip()
    else:
        result = name + "_" + suffix.strip()
    return result.replace(" ", "_")


def python_class_name(name: str) -> str:
    """Generate a valid python variable name."""
    return name.replace(" ", "")


def float_or_none(value: str) -> float | None:
    """Calculate a float value or none."""
    if value == "":
        return None
    return float(value)


def get_base_address(registers: list) -> int:
    """Get the base address from a register block."""
    result = 35000
    for register in registers:
        result = min(result, int(register[0]))
    return result - 1


def get_block_len(registers: list) -> int:
    """Get the base address from a register block."""
    base = 35000
    highest = 0
    for register in registers:
        base = min(base, int(register[0]))
        highest = max(highest, int(register[0]))
    return highest - base + 1


def generate_heatpump(
    root_path: Path,
    template: Template,
    modbus_files: list[ModbusFile],
    heatpump_type: str,
):
    """Generate the python file for a heat pump."""
    register_blocks = []

    for modbus_file in modbus_files:
        with modbus_file.path.open("r") as f:
            reader = csv.reader(f)
            data = list(reader)

        api_data = data[1:]
        register_blocks.append(
            {
                "name": modbus_file.name,
                "count": get_block_len(api_data),
                "base_address": get_base_address(api_data),
                "register_block": api_data,
                "register_type": modbus_file.register_type,
            }
        )

    generated_content = template.render(registers=register_blocks, heatpump_type=heatpump_type)

    generated_path = root_path / f"pystiebeleltron/{heatpump_type.lower()}.py"

    with generated_path.open("w") as f:
        f.write(generated_content)


def main() -> None:
    """Generate the python files."""
    root = Path.cwd()
    scripts_path = Path(__file__).parent

    env = Environment(
        loader=FileSystemLoader(str(scripts_path / "templates")),
        autoescape=select_autoescape(),
    )
    env.globals.update(python_name=python_name, python_class_name=python_class_name, float_or_none=float_or_none, ENERGY_DATA_BLOCK_NAME=ENERGY_DATA_BLOCK_NAME)

    wpm_template = env.get_template("wpmtemplate.j2")
    wpm_modbus_files = [
        ModbusFile(
            name="System Values",
            path=root / "api/wpm_system_values.csv",
            register_type=RegisterType.INPUT_REGISTER,
        ),
        ModbusFile(
            name="System Parameters",
            path=root / "api/wpm_system_parameters.csv",
            register_type=RegisterType.HOLDING_REGISTER,
        ),
        ModbusFile(
            name="System State",
            path=root / "api/wpm_system_state.csv",
            register_type=RegisterType.INPUT_REGISTER,
        ),
        ModbusFile(
            name="Energy Data",
            path=root / "api/wpm_energy_data.csv",
            register_type=RegisterType.INPUT_REGISTER,
        ),
    ]
    generate_heatpump(root, wpm_template, wpm_modbus_files, "Wpm")

    lwz_template = env.get_template("lwztemplate.j2")
    lwz_modbus_files = [
        ModbusFile(
            name="System Values",
            path=root / "api/lwz_system_values.csv",
            register_type=RegisterType.INPUT_REGISTER,
        ),
        ModbusFile(
            name="System Parameters",
            path=root / "api/lwz_system_parameters.csv",
            register_type=RegisterType.HOLDING_REGISTER,
        ),
        ModbusFile(
            name="System State",
            path=root / "api/lwz_system_state.csv",
            register_type=RegisterType.INPUT_REGISTER,
        ),
        ModbusFile(
            name="Energy Data",
            path=root / "api/lwz_energy_data.csv",
            register_type=RegisterType.INPUT_REGISTER,
        ),
    ]
    generate_heatpump(root, lwz_template, lwz_modbus_files, "Lwz")


if __name__ == "__main__":
    main()
    print("Done!")



import pytest
from pystiebeleltron.lwz import LwzEnergyDataRegisters, LwzStiebelEltronAPI, LwzSystemValuesRegisters, OperatingMode
from pystiebeleltron.wpm import WpmEnergyDataRegisters, WpmStiebelEltronAPI, WpmSystemValuesRegisters
from pymodbus.pdu.register_message import (
    ReadInputRegistersResponse,
)


async def read_registers(client, address: int, *, count: int = 1, slave: int = 0, no_response_expected: bool = False) -> ReadInputRegistersResponse:
    """Read a slice from the input register."""
    return ReadInputRegistersResponse(address=address, count=count, registers=list(range(count)))


@pytest.mark.asyncio()
async def test_wpm(mocker):
    api = WpmStiebelEltronAPI("localhost")
    mock_connect = mocker.patch("pymodbus.client.AsyncModbusTcpClient.connect")
    mock_close = mocker.patch("pymodbus.client.AsyncModbusTcpClient.close")
    mocker.patch("pymodbus.client.AsyncModbusTcpClient.read_holding_registers", read_registers)
    mocker.patch("pymodbus.client.AsyncModbusTcpClient.read_input_registers", read_registers)

    await api.connect()
    mock_connect.assert_called_once()

    await api.async_update()

    assert api.get_register_value(WpmSystemValuesRegisters.ACTUAL_TEMPERATURE_FEK) == 0.2

    assert api.get_register_value(WpmEnergyDataRegisters.VD_HEATING_DAY_AND_TOTAL_CONSUMED) == 12021

    await api.close()
    mock_close.assert_called_once()


@pytest.mark.asyncio()
async def test_lwz(mocker):
    api = LwzStiebelEltronAPI("localhost")
    mock_connect = mocker.patch("pymodbus.client.AsyncModbusTcpClient.connect")
    mock_close = mocker.patch("pymodbus.client.AsyncModbusTcpClient.close")
    mocker.patch("pymodbus.client.AsyncModbusTcpClient.read_holding_registers", read_registers)
    mocker.patch("pymodbus.client.AsyncModbusTcpClient.read_input_registers", read_registers)

    await api.connect()
    mock_connect.assert_called_once()

    await api.async_update()

    assert api.get_register_value(LwzSystemValuesRegisters.RELATIVE_HUMIDITY_HC1) == 0.2

    assert api.get_register_value(LwzEnergyDataRegisters.HEAT_METER_HTG_DAY_AND_TOTAL) == 2001

    assert api.get_current_humidity() == 0.2
    assert api.get_current_temp() == 0.0
    assert api.get_target_temp() == 0.1

    assert api.get_operation() == OperatingMode.EMERGENCY_OPERATION

    assert api.get_register_value(LwzSystemValuesRegisters.COMPRESSOR_STARTS) == 30033

    await api.close()
    mock_close.assert_called_once()



#!/usr/bin/env python3
import asyncio
from pystiebeleltron.wpm import WpmStiebelEltronAPI, WpmSystemParametersRegisters, WpmSystemValuesRegisters

host_ip = "192.168.1.20"


async def main():
    api = WpmStiebelEltronAPI(host_ip, 502)
    await api.connect()

    await api.async_update()

    for k, v in api._data.items():
        if v is not None:
            print(f"{k.name} ({k.value}): {v}")

    outside_temp = api.get_register_value(WpmSystemValuesRegisters.OUTSIDE_TEMPERATURE)
    print(f"The current outside temperature is {outside_temp} °C")

    comfort_temp = api.get_register_value(WpmSystemParametersRegisters.COMFORT_TEMPERATURE)
    print(f"The current water comfort temperature is {comfort_temp} °C")

    # Test set_target_temp
    print("Setting temperature to 50.0")
    await api.write_register_value(WpmSystemParametersRegisters.COMFORT_TEMPERATURE, 50)
    await asyncio.sleep(3)
    await api.async_update()
    mod_temp = api.get_register_value(WpmSystemParametersRegisters.COMFORT_TEMPERATURE)
    if mod_temp != 50.0:
        print("setting the water comfort temperature failed!")
    if mod_temp != comfort_temp:
        await api.write_register_value(WpmSystemParametersRegisters.COMFORT_TEMPERATURE, comfort_temp)
        await asyncio.sleep(3)
        await api.async_update()
    print(f"get_target_temp: {api.get_register_value(WpmSystemParametersRegisters.COMFORT_TEMPERATURE)}")
    await api.close()


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    loop.run_until_complete(main())



#!/bin/env python
"""
Pymodbus Asynchronous Server Example
--------------------------------------------------------------------------

The asynchronous server is a high performance implementation using the
twisted library as its backend.  This allows it to scale to many thousands
of nodes which can be helpful for testing monitoring software.
"""
# --------------------------------------------------------------------------- #
# import the various server implementations
# --------------------------------------------------------------------------- #
from pymodbus.server import StartTcpServer, ServerStop

from pymodbus.device import ModbusDeviceIdentification
from pymodbus.datastore import ModbusSequentialDataBlock
from pymodbus.datastore import ModbusSlaveContext, ModbusServerContext


class MockModbusServer(object):
    # --------------------------------------------------------------------------- #
    # configure the service logging
    # --------------------------------------------------------------------------- #
    import logging
    FORMAT = ('%(asctime)-15s %(threadName)-15s'
              ' %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s')
    logging.basicConfig(format=FORMAT)
    log = logging.getLogger()
    log.setLevel(logging.DEBUG)

    def run_async_server(self):
        # ----------------------------------------------------------------------- #
        # initialize your data store
        # ----------------------------------------------------------------------- #
        # The datastores only respond to the addresses that they are initialized to
        # Therefore, if you initialize a DataBlock to addresses from 0x00 to 0xFF,
        # a request to 0x100 will respond with an invalid address exception.
        # This is because many devices exhibit this kind of behavior (but not all)
        #
        #     block = ModbusSequentialDataBlock(0x00, [0]*0xff)
        #
        # Continuing, you can choose to use a sequential or a sparse DataBlock in
        # your data context.  The difference is that the sequential has no gaps in
        # the data while the sparse can. Once again, there are devices that exhibit
        # both forms of behavior::
        #
        #     block = ModbusSparseDataBlock({0x00: 0, 0x05: 1})
        #     block = ModbusSequentialDataBlock(0x00, [0]*5)
        #
        # Alternately, you can use the factory methods to initialize the DataBlocks
        # or simply do not pass them to have them initialized to 0x00 on the full
        # address range::
        #
        #     store = ModbusSlaveContext(di = ModbusSequentialDataBlock.create())
        #     store = ModbusSlaveContext()
        #
        # Finally, you are allowed to use the same DataBlock reference for every
        # table or you you may use a seperate DataBlock for each table.
        # This depends if you would like functions to be able to access and modify
        # the same data or not::
        #
        #     block = ModbusSequentialDataBlock(0x00, [0]*0xff)
        #     store = ModbusSlaveContext(di=block, co=block, hr=block, ir=block)
        #
        # The server then makes use of a server context that allows the server to
        # respond with different slave contexts for different unit ids. By default
        # it will return the same context for every unit id supplied (broadcast
        # mode).
        # However, this can be overloaded by setting the single flag to False
        # and then supplying a dictionary of unit id to context mapping::
        #
        #     slaves  = {
        #         0x01: ModbusSlaveContext(...),
        #         0x02: ModbusSlaveContext(...),
        #         0x03: ModbusSlaveContext(...),
        #     }
        #     context = ModbusServerContext(slaves=slaves, single=False)
        #
        # The slave context can also be initialized in zero_mode which means that a
        # request to address(0-7) will map to the address (0-7). The default is
        # False which is based on section 4.4 of the specification, so address(0-7)
        # will map to (1-8)::
        #
        #     store = ModbusSlaveContext(..., zero_mode=True)
        # ----------------------------------------------------------------------- #
        store = ModbusSlaveContext(
            hr=ModbusSequentialDataBlock(0, [0]*3000),
            ir=ModbusSequentialDataBlock(0, [0]*3000))
        self.context = ModbusServerContext(slaves=store, single=True)

        # ----------------------------------------------------------------------- #
        # initialize the server information
        # ----------------------------------------------------------------------- #
        # If you don't set this or any fields, they are defaulted to empty strings.
        # ----------------------------------------------------------------------- #
        identity = ModbusDeviceIdentification()
        identity.VendorName = 'Pymodbus'
        identity.ProductCode = 'PM'
        identity.VendorUrl = 'http://github.com/bashwork/pymodbus/'
        identity.ProductName = 'Pymodbus Server'
        identity.ModelName = 'Pymodbus Server'
        identity.MajorMinorRevision = '1.5'

        # ----------------------------------------------------------------------- #
        # run the server you want
        # ----------------------------------------------------------------------- #

        # TCP Server
        StartTcpServer(self.context, identity=identity, address=("localhost", 5020))

    def stop_async_server(self):
        ServerStop()

    def update_context(self, register, address, values):
        """ Update values of the active context. It should be noted
        that there is a race condition for the update.

        :param register: Type of register to update,
                            3: holding register
                            4: input register
        :param address: The starting address of the value to be changed
        :param values: List of values
        """
        assert register == 3 or register == 4
        slave_id = 0x00
        old_values = self.context[slave_id].getValues(register,
                                                      address, count=1)
        self.log.debug("Change value at address {} from {} to {}".format(
            address, old_values, values))
        self.context[slave_id].setValues(register, address, values)

    def update_holding_register(self, address, value):
        """ Update value of a holding register.

        :param address: Address to update
        :param value: Value to save
        """
        self.log.debug("Update holding register: {}:{}".format(address,
                                                               int(value)))
        self.update_context(3, address, [int(value)])

    def update_input_register(self, address, value):
        """ Update value of an input register.

        :param address: Address to update
        :param value: Value to save
        """
        self.log.debug("Update input register: {}:{}".format(address,
                                                             int(value)))
        self.update_context(4, address, [int(value)])


if __name__ == "__main__":
    mms = MockModbusServer()
    mms.run_async_server()



#!/usr/bin/env python
import os
import time
import pytest

# Import modbus mockup server requirements
from threading import Thread
from test.mock_modbus_server import MockModbusServer as ModbusServer

# Import client requirementss
from pymodbus.client import ModbusTcpClient as ModbusClient
from pystiebeleltron import pystiebeleltron as pyse

# Modbus server connection details
host_ip = "127.0.0.1" #"192.168.1.20"
host_port = 5020 #502
slave = 1


class TestStiebelEltronApi:
    #__slots__ = 'api'

    @pytest.fixture(scope="module")
    def pymb_s(self, request):
        mb_s = ModbusServer()

        # Cleanup after last test did run (will run as well, if something fails in setup).
        def fin():
            stop_thread = Thread(target=mb_s.stop_async_server,
                                 name="StopReactor")
            stop_thread.start()
            if stop_thread.is_alive():
                stop_thread.join()

            if mms_thread.is_alive():
                mms_thread.join()

        request.addfinalizer(fin)

        # Start Mock Modbus server
        mms_thread = Thread(target=mb_s.run_async_server,
                            name="MockModbusServer")
        mms_thread.start()
        time.sleep(0.1)

        return mb_s

    @pytest.fixture(scope="module")
    def pyse_api(self, request, pymb_s):
        # parameter pymb_s leads to call of fixture
        mb_c = ModbusClient(host=host_ip, port=host_port, timeout=2)
        api = pyse.StiebelEltronAPI(mb_c, slave, update_on_read=True)

        # Cleanup after last test (will run as well, if setup fails).
        def fin():
            mb_c.close()
            time.sleep(0.5)

        request.addfinalizer(fin)

        # Connect Modbus client
        connected = mb_c.connect()
        assert connected

        # Read values from device (server)
        success = api.update()
        assert success

        return api

    def test_temperature_read(self, pyse_api, pymb_s):
        pymb_s.update_input_register(0, 21.5*10)
        assert pyse_api.get_current_temp() == 21.5

        pymb_s.update_holding_register(1001, 22.5*10)
        assert pyse_api.get_target_temp() == 22.5

    def test_temperature_write(self, pyse_api):
        temperature = 22.5
        pyse_api.set_target_temp(temperature)
        time.sleep(3)

        assert pyse_api.get_target_temp() == temperature

    def test_operation(self, pyse_api):
        operation = 'DHW'
        pyse_api.set_operation(operation)
        time.sleep(3)

        assert pyse_api.get_operation() == operation

    def test_humidity(self, pyse_api, pymb_s):
        humidity = 49.5
        pymb_s.update_input_register(2, humidity*10)
        assert pyse_api.get_current_humidity() == humidity

    def test_statuses(self, pyse_api, pymb_s):
        pymb_s.update_input_register(2000, 0x0004)
        assert pyse_api.get_heating_status() is True
        assert pyse_api.get_cooling_status() is False
        assert pyse_api.get_filter_alarm_status() is False

        pymb_s.update_input_register(2000, 0x0008)
        assert pyse_api.get_heating_status() is False
        assert pyse_api.get_cooling_status() is True
        assert pyse_api.get_filter_alarm_status() is False

        pymb_s.update_input_register(2000, 0x2100)
        assert pyse_api.get_heating_status() is False
        assert pyse_api.get_cooling_status() is False
        assert pyse_api.get_filter_alarm_status() is True



#!/usr/bin/env python3
import time
from pystiebeleltron import pystiebeleltron as pyse
from pymodbus.client import ModbusTcpClient as ModbusClient

host_ip = "192.168.1.20"
host_port = 502
slave = 1

def test_function(mod, fun):
    """Executes the given function on the Stiebel Heatpump and prints the result."""
    result = getattr(mod, fun) # Executes function directly, instead of giving back the function reference
    print("{}: {}".format(str(fun), str(result)))


def execute_tests(unit):
    """Execute the provided tests."""
    test_function(unit, "get_current_temp")
    test_function(unit, "get_current_humidity")
    test_function(unit, "get_target_temp")
    test_function(unit, "get_operation")
    test_function(unit, "get_filter_alarm_status")
    test_function(unit, "get_heating_status")
    test_function(unit, "get_cooling_status")

    # Test set_target_temp
    print("Setting temperature to 20.0")
    current_temp = unit.get_target_temp
    unit.set_target_temp(20.0)
    time.sleep(3)
    unit.update()
    mod_temp = unit.get_target_temp
    if mod_temp != 20.0:
        print("unit.set_target_temp failed!")
    if mod_temp != current_temp:
        unit.set_target_temp(current_temp)
        time.sleep(3)
        unit.update()
    print("get_target_temp: {}".format(unit.get_target_temp))


def main():
    client = ModbusClient(host=host_ip,
                          port=host_port,
                          timeout=2)
    client.connect()

    unit = pyse.StiebelEltronAPI(client, slave)
    unit.update()

    execute_tests(unit)

    client.close()

if __name__ == "__main__":
    main()



"""
Connection to a Stiebel Eltron ModBus API.

See API details:
https://www.stiebel-eltron.de/content/dam/ste/de/de/home/services/Downloadlisten/ISG%20Modbus_Stiebel_Bedienungsanleitung.pdf

Types of data:

Data | Value      | Multiplier  | Multiplier  | Signed | Step   | Step
type | range      | for reading | for writing |        | size 1 | size 5
-----|------------|-------------|-------------|--------|--------|-------
2    | -3276.8 to | 0.1         | 10          | Yes    | 0.1    | 0.5
     |  3276.7    |             |             |        |        |
6    | 0 to 65535 | 1           | 1           | No     | 1      | 5
7    | -327.68 to | 0.01        | 100         | Yes    | 0.01   | 0.05
     |  327.67    |             |             |        |        |
8    | 0 to 255   | 1           | 1           | No     | 1      | 5
"""

from pymodbus.client.mixin import ModbusClientMixin

# Error - sensor lead is missing or disconnected.
ERROR_NOTAVAILABLE = -60
# Error - short circuit of the sensor lead.
ERROR_SHORTCUT = -50
# Error - object unavailable.
ERROR_OBJ_UNAVAILBLE = 0x8000

UNAVAILABLE_OBJECT = 32768

# Block 1 System values (Read input register) - page 29
B1_START_ADDR = 0

B1_REGMAP_INPUT = {
    # HC = Heating Circuit
    'ACTUAL_ROOM_TEMPERATURE_HC1':      {'addr':  0, 'type': 2, 'value': 0},
    'SET_ROOM_TEMPERATURE_HC1':         {'addr':  1, 'type': 2, 'value': 0},
    'RELATIVE_HUMIDITY_HC1':            {'addr':  2, 'type': 2, 'value': 0},
    'ACTUAL_ROOM_TEMPERATURE_HC2':      {'addr':  3, 'type': 2, 'value': 0},
    'SET_ROOM_TEMPERATURE_HC2':         {'addr':  4, 'type': 2, 'value': 0},
    'RELATIVE_HUMIDITY_HC2':            {'addr':  5, 'type': 2, 'value': 0},
    'OUTSIDE_TEMPERATURE':              {'addr':  6, 'type': 2, 'value': 0},
    'ACTUAL_VALUE_HC1':                 {'addr':  7, 'type': 2, 'value': 0},
    'SET_VALUE_HC1':                    {'addr':  8, 'type': 2, 'value': 0},
    'ACTUAL_VALUE_HC2':                 {'addr':  9, 'type': 2, 'value': 0},
    'SET_VALUE_HC2':                    {'addr': 10, 'type': 2, 'value': 0},
    'FLOW_TEMPERATURE':                 {'addr': 11, 'type': 2, 'value': 0},
    'RETURN_TEMPERATURE':               {'addr': 12, 'type': 2, 'value': 0},
    'PRESSURE_HEATING_CIRCUIT':         {'addr': 13, 'type': 2, 'value': 0},
    'FLOW_RATE':                        {'addr': 14, 'type': 2, 'value': 0},
    'ACTUAL_DHW_TEMPERATURE':           {'addr': 15, 'type': 2, 'value': 0},
    'SET_DHW_TEMPERATURE':              {'addr': 16, 'type': 2, 'value': 0},
    'VENTILATION_AIR_ACTUAL_FAN_SPEED': {'addr': 17, 'type': 6, 'value': 0},
    'VENTILATION_AIR_SET_FLOW_RATE':    {'addr': 18, 'type': 6, 'value': 0},
    'EXTRACT_AIR_ACTUAL_FAN_SPEED':     {'addr': 19, 'type': 6, 'value': 0},
    'EXTRACT_AIR_SET_FLOW_RATE':        {'addr': 20, 'type': 6, 'value': 0},
    'EXTRACT_AIR_HUMIDITY':             {'addr': 21, 'type': 6, 'value': 0},
    'EXTRACT_AIR_TEMPERATURE':          {'addr': 22, 'type': 2, 'value': 0},
    'EXTRACT_AIR_DEW_POINT':            {'addr': 23, 'type': 2, 'value': 0},
    'DEW_POINT_TEMPERATUR_HC1':         {'addr': 24, 'type': 2, 'value': 0},
    'DEW_POINT_TEMPERATUR_HC2':         {'addr': 25, 'type': 2, 'value': 0},
    'COLLECTOR_TEMPERATURE':            {'addr': 26, 'type': 2, 'value': 0},
    'HOT_GAS_TEMPERATURE':              {'addr': 27, 'type': 2, 'value': 0},
    'HIGH_PRESSURE':                    {'addr': 28, 'type': 7, 'value': 0},
    'LOW_PRESSURE':                     {'addr': 29, 'type': 7, 'value': 0},
    'COMPRESSOR_STARTS':                {'addr': 30, 'type': 6, 'value': 0},
    'COMPRESSOR_SPEED':                 {'addr': 31, 'type': 2, 'value': 0},
    'MIXED_WATER_AMOUNT':               {'addr': 32, 'type': 6, 'value': 0}
}


# Block 2 System parameters (Read/write holding register) - page 30
B2_START_ADDR = 1000

B2_REGMAP_HOLDING = {
    'OPERATING_MODE':           {'addr': 1000, 'type': 8, 'value': 0},
    'ROOM_TEMP_HEAT_DAY_HC1':   {'addr': 1001, 'type': 2, 'value': 0},
    'ROOM_TEMP_HEAT_NIGHT_HC1': {'addr': 1002, 'type': 2, 'value': 0},
    'MANUAL_SET_TEMP_HC1':      {'addr': 1003, 'type': 2, 'value': 0},
    'ROOM_TEMP_HEAT_DAY_HC2':   {'addr': 1004, 'type': 2, 'value': 0},
    'ROOM_TEMP_HEAT_NIGHT_HC2': {'addr': 1005, 'type': 2, 'value': 0},
    'MANUAL_SET_TEAMP_HC2':     {'addr': 1006, 'type': 2, 'value': 0},
    'GRADIENT_HC1':             {'addr': 1007, 'type': 7, 'value': 0},
    'LOW_END_HC1':              {'addr': 1008, 'type': 2, 'value': 0},
    'GRADIENT_HC2':             {'addr': 1009, 'type': 7, 'value': 0},
    'LOW_END_HC2':              {'addr': 1010, 'type': 2, 'value': 0},
    'DHW_TEMP_SET_DAY':         {'addr': 1011, 'type': 2, 'value': 0},
    'DHW_TEMP_SET_NIGHT':       {'addr': 1012, 'type': 2, 'value': 0},
    'DHW_TEMP_SET_MANUAL':      {'addr': 1013, 'type': 2, 'value': 0},
    'MWM_SET_DAY':              {'addr': 1014, 'type': 6, 'value': 0},
    'MWM_SET_NIGHT':            {'addr': 1015, 'type': 6, 'value': 0},
    'MWM_SET_MANUAL':           {'addr': 1016, 'type': 6, 'value': 0},
    'DAY_STAGE':                {'addr': 1017, 'type': 6, 'value': 0},
    'NIGHT_STAGE':              {'addr': 1018, 'type': 6, 'value': 0},
    'PARTY_STAGE':              {'addr': 1019, 'type': 6, 'value': 0},
    'MANUAL_STAGE':             {'addr': 1020, 'type': 6, 'value': 0},
    'ROOM_TEMP_COOL_DAY_HC1':   {'addr': 1021, 'type': 2, 'value': 0},
    'ROOM_TEMP_COOL_NIGHT_HC1': {'addr': 1022, 'type': 2, 'value': 0},
    'ROOM_TEMP_COOL_DAY_HC2':   {'addr': 1023, 'type': 2, 'value': 0},
    'ROOM_TEMP_COOL_NIGHT_HC2': {'addr': 1024, 'type': 2, 'value': 0},
    'RESET':                    {'addr': 1025, 'type': 6, 'value': 0},
    'RESTART_ISG':              {'addr': 1026, 'type': 6, 'value': 0}
}

B2_OPERATING_MODE_READ = {
    # AUTOMATIK
    11: 'AUTOMATIC',
    # BEREITSCHAFT
    1: 'STANDBY',
    # TAGBETRIEB
    3: 'DAY MODE',
    # ABSENKBETRIEB
    4: 'SETBACK MODE',
    # WARMWASSER
    5: 'DHW',
    # HANDBETRIEB
    14: 'MANUAL MODE',
    # NOTBETRIEB
    0: 'EMERGENCY OPERATION'
}

B2_OPERATING_MODE_WRITE = {
    value: key for key, value in B2_OPERATING_MODE_READ.items()
}

B2_RESET = {
    'OFF': 0,
    'ON': 1
}

B2_RESTART_ISG = {
    'OFF': 0,
    'RESET': 1,
    'MENU': 2
}

# Block 3 System status (Read input register) - page 31
B3_START_ADDR = 2000

B3_REGMAP_INPUT = {
    'OPERATING_STATUS': {'addr': 2000, 'type': 6, 'value': 0},
    'FAULT_STATUS':     {'addr': 2001, 'type': 6, 'value': 0},
    'BUS_STATUS':       {'addr': 2002, 'type': 6, 'value': 0}
}

B3_OPERATING_STATUS = {
    'SWITCHING_PROGRAM_ENABLED': (1 << 0),
    'COMPRESSOR': (1 << 1),
    'HEATING': (1 << 2),
    'COOLING': (1 << 3),
    'DHW': (1 << 4),
    'ELECTRIC_REHEATING': (1 << 5),
    'SERVICE': (1 << 6),
    'POWER-OFF': (1 << 7),
    'FILTER': (1 << 8),
    'VENTILATION': (1 << 9),
    'HEATING_CIRCUIT_PUMP': (1 << 10),
    'EVAPORATOR_DEFROST': (1 << 11),
    'FILTER_EXTRACT_AIR': (1 << 12),
    'FILTER_VENTILATION_AIR': (1 << 13),
    'HEAT-UP_PROGRAM': (1 << 14)
}

B3_FAULT_STATUS = {
    'NO_FAULT': 0,
    'FAULT': 1
}

B3_BUS_STATUS = {
    'STATUS OK': 0,
    'STATUS ERROR': -1,
    'ERROR-PASSIVE': -2,
    'BUS-OFF': -3,
    'PHYSICAL-ERROR': -4
}


class StiebelEltronAPI():
    """Stiebel Eltron API."""

    def __init__(self, conn: ModbusClientMixin, slave=1, update_on_read=False):
        """Initialize Stiebel Eltron communication."""
        self._conn = conn
        self._block_1_input_regs = B1_REGMAP_INPUT
        self._block_2_holding_regs = B2_REGMAP_HOLDING
        self._block_3_input_regs = B3_REGMAP_INPUT
        self._slave = slave
        self._update_on_read = update_on_read

    def update(self):
        """Request current values from heat pump."""
        ret = True
        try:
            block_1_result_input = self._conn.read_input_registers(
                slave=self._slave,
                address=B1_START_ADDR,
                count=len(self._block_1_input_regs)).registers
            block_2_result_holding = self._conn.read_holding_registers(
                slave=self._slave,
                address=B2_START_ADDR,
                count=len(self._block_2_holding_regs)).registers
            block_3_result_input = self._conn.read_input_registers(
                slave=self._slave,
                address=B3_START_ADDR,
                count=len(self._block_3_input_regs)).registers
        except AttributeError:
            # The unit does not reply reliably
            ret = False
            print("Modbus read failed")
        else:
            for v in self._block_1_input_regs.values():
                v['value'] = block_1_result_input[v['addr'] - B1_START_ADDR]

            for v in self._block_2_holding_regs.values():
                v['value'] = block_2_result_holding[v['addr'] - B2_START_ADDR]

            for v in self._block_3_input_regs.values():
                v['value'] = block_3_result_input[v['addr'] - B3_START_ADDR]

        return ret

    def get_conv_val(self, name: str):
        """Read and convert value.

        Args:
            name: Name of value to be read.

        Returns:
            Actual value or None.
        """
        value_entry = self._block_1_input_regs.get(name)
        if value_entry is None:
            value_entry = self._block_2_holding_regs.get(name)
        if value_entry is None:
            value_entry = self._block_3_input_regs.get(name)
        if value_entry is None:
            return None

        if value_entry['type'] == 2:
            return value_entry['value'] * 0.1
        if value_entry['type'] == 7:
            return value_entry['value'] * 0.01

        return value_entry['value']

#    def get_raw_input_register(self, name):
#        """Get raw register value by name."""
#        if self._update_on_read:
#            self.update()
#        return self._block_1_input_regs[name]

#    def get_raw_holding_register(self, name):
#        """Get raw register value by name."""
#        if self._update_on_read:
#            self.update()
#        return self._block_2_holding_regs[name]

#    def set_raw_holding_register(self, name, value):
#        """Write to register by name."""
#        self._conn.write_register(
#            slave=self._slave,
#            address=(self._holding_regs[name]['addr']),
#            value=value)

    # Handle room temperature & humidity

    def get_current_temp(self):
        """Get the current room temperature."""
        if self._update_on_read:
            self.update()
        return self.get_conv_val('ACTUAL_ROOM_TEMPERATURE_HC1')

    def get_target_temp(self):
        """Get the target room temperature."""
        if self._update_on_read:
            self.update()
        return self.get_conv_val('ROOM_TEMP_HEAT_DAY_HC1')

    def set_target_temp(self, temp: float):
        """Set the target room temperature (day)(HC1)."""
        self._conn.write_register(
            slave=self._slave,
            address=(
                self._block_2_holding_regs['ROOM_TEMP_HEAT_DAY_HC1']['addr']),
            value=round(temp * 10.0))

    def get_current_humidity(self):
        """Get the current room humidity."""
        if self._update_on_read:
            self.update()
        return self.get_conv_val('RELATIVE_HUMIDITY_HC1')

    # Handle operation mode

    def get_operation(self):
        """Return the current mode of operation."""
        if self._update_on_read:
            self.update()

        op_mode = self.get_conv_val('OPERATING_MODE')
        return B2_OPERATING_MODE_READ.get(op_mode, 'UNKNOWN')

    def set_operation(self, mode: str):
        """Set the operation mode."""
        self._conn.write_register(
            slave=self._slave,
            address=(self._block_2_holding_regs['OPERATING_MODE']['addr']),
            value=B2_OPERATING_MODE_WRITE.get(mode))

    # Handle device status

    def get_heating_status(self):
        """Return heater status."""
        if self._update_on_read:
            self.update()
        return bool(self.get_conv_val('OPERATING_STATUS') &
                    B3_OPERATING_STATUS['HEATING'])

    def get_cooling_status(self):
        """Cooling status."""
        if self._update_on_read:
            self.update()
        return bool(self.get_conv_val('OPERATING_STATUS') &
                    B3_OPERATING_STATUS['COOLING'])

    def get_filter_alarm_status(self):
        """Return filter alarm."""
        if self._update_on_read:
            self.update()

        filter_mask = (B3_OPERATING_STATUS['FILTER'] |
                       B3_OPERATING_STATUS['FILTER_EXTRACT_AIR'] |
                       B3_OPERATING_STATUS['FILTER_VENTILATION_AIR'])
        return bool(self.get_conv_val('OPERATING_STATUS') & filter_mask)
